@@ -3,7 +3,7 @@ import json
 from django import urls as urlresolvers
 from django.conf import settings
 from django.urls.exceptions import NoReverseMatch
-from django.utils.html import format_html
+from django.utils.html import format_html, format_html_join
 from django.utils.safestring import mark_safe
 
 from auditlog.models import LogEntry
@@ -63,16 +63,61 @@ class LogEntryAdminMixin:
         if obj.action == LogEntry.Action.DELETE:
             return ""  # delete
         changes = json.loads(obj.changes)
-        msg = "<table><tr><th>#</th><th>Field</th><th>From</th><th>To</th></tr>"
-        for i, field in enumerate(sorted(changes), 1):
-            value = [i, field] + (
-                ["***", "***"] if field == "password" else changes[field]
-            )
-            msg += format_html(
-                "<tr><td>{}</td><td>{}</td><td>{}</td><td>{}</td></tr>", *value
-            )
 
-        msg += "</table>"
-        return mark_safe(msg)
+        atom_changes = {}
+        m2m_changes = {}
+
+        for field, change in changes.items():
+            if isinstance(change, dict):
+                assert (
+                    change["type"] == "m2m"
+                ), "Only m2m operations are expected to produce dict changes now"
+                m2m_changes[field] = change
+            else:
+                atom_changes[field] = change
+
+        msg = []
+
+        if atom_changes:
+            msg.append("<table>")
+            msg.append(self._format_header("#", "Field", "From", "To"))
+            for i, (field, change) in enumerate(sorted(atom_changes.items()), 1):
+                value = [i, field] + (["***", "***"] if field == "password" else change)
+                msg.append(self._format_line(*value))
+            msg.append("</table>")
+
+        if m2m_changes:
+            msg.append("<table>")
+            msg.append(self._format_header("#", "Relationship", "Action", "Objects"))
+            for i, (field, change) in enumerate(sorted(m2m_changes.items()), 1):
+                change_html = format_html_join(
+                    mark_safe("<br>"),
+                    "{}",
+                    [(value,) for value in change["objects"]],
+                )
+
+                msg.append(
+                    format_html(
+                        "<tr><td>{}</td><td>{}</td><td>{}</td><td>{}</td></tr>",
+                        i,
+                        field,
+                        change["operation"],
+                        change_html,
+                    )
+                )
+
+            msg.append("</table>")
+
+        return mark_safe("".join(msg))
 
     msg.short_description = "Changes"
+
+    def _format_header(self, *labels):
+        return format_html(
+            "".join(["<tr>", "<th>{}</th>" * len(labels), "</tr>"]), *labels
+        )
+
+    def _format_line(self, *values):
+        return format_html(
+            "".join(["<tr>", "<td>{}</td>" * len(values), "</tr>"]), *values
+        )
