@@ -21,8 +21,8 @@ class LogEntryManager(models.Manager):
 
     def log_create(self, instance, **kwargs):
         """
-        Helper method to create a new log entry. This method automatically populates some fields when no explicit value
-        is given.
+        Helper method to create a new log entry. This method automatically populates some fields when no
+        explicit value is given.
 
         :param instance: The model instance to log a change for.
         :type instance: Model
@@ -47,8 +47,8 @@ class LogEntryManager(models.Manager):
             if callable(get_additional_data):
                 kwargs.setdefault("additional_data", get_additional_data())
 
-            # Delete log entries with the same pk as a newly created model. This should only be necessary when an pk is
-            # used twice.
+            # Delete log entries with the same pk as a newly created model.
+            # This should only be necessary when an pk is used twice.
             if kwargs.get("action", None) is LogEntry.Action.CREATE:
                 if (
                     kwargs.get("object_id", None) is not None
@@ -69,6 +69,54 @@ class LogEntryManager(models.Manager):
             return self.create(**kwargs)
         return None
 
+    def log_m2m_changes(
+        self, changed_queryset, instance, operation, field_name, **kwargs
+    ):
+        """Create a new "changed" log entry from m2m record.
+
+        :param changed_queryset: The added or removed related objects.
+        :type changed_queryset: QuerySet
+        :param instance: The model instance to log a change for.
+        :type instance: Model
+        :param operation: "add" or "delete".
+        :type action: str
+        :param field_name: The name of the changed m2m field.
+        :type field_name: str
+        :param kwargs: Field overrides for the :py:class:`LogEntry` object.
+        :return: The new log entry or `None` if there were no changes.
+        :rtype: LogEntry
+        """
+
+        pk = self._get_pk_value(instance)
+        if changed_queryset is not None:
+            kwargs.setdefault(
+                "content_type", ContentType.objects.get_for_model(instance)
+            )
+            kwargs.setdefault("object_pk", pk)
+            kwargs.setdefault("object_repr", smart_str(instance))
+            kwargs.setdefault("action", LogEntry.Action.UPDATE)
+
+            if isinstance(pk, int):
+                kwargs.setdefault("object_id", pk)
+
+            get_additional_data = getattr(instance, "get_additional_data", None)
+            if callable(get_additional_data):
+                kwargs.setdefault("additional_data", get_additional_data())
+
+            objects = [smart_str(instance) for instance in changed_queryset]
+            kwargs["changes"] = json.dumps(
+                {
+                    field_name: {
+                        "type": "m2m",
+                        "operation": operation,
+                        "objects": objects,
+                    }
+                }
+            )
+            return self.create(**kwargs)
+
+        return None
+
     def get_for_object(self, instance):
         """
         Get log entries for the specified model instance.
@@ -163,21 +211,21 @@ class LogEntryManager(models.Manager):
 
 class LogEntry(models.Model):
     """
-    Represents an entry in the audit log. The content type is saved along with the textual and numeric (if available)
-    primary key, as well as the textual representation of the object when it was saved. It holds the action performed
-    and the fields that were changed in the transaction.
+    Represents an entry in the audit log. The content type is saved along with the textual and numeric
+    (if available) primary key, as well as the textual representation of the object when it was saved.
+    It holds the action performed and the fields that were changed in the transaction.
 
-    If AuditlogMiddleware is used, the actor will be set automatically. Keep in mind that editing / re-saving LogEntry
-    instances may set the actor to a wrong value - editing LogEntry instances is not recommended (and it should not be
-    necessary).
+    If AuditlogMiddleware is used, the actor will be set automatically. Keep in mind that
+    editing / re-saving LogEntry instances may set the actor to a wrong value - editing LogEntry
+    instances is not recommended (and it should not be necessary).
     """
 
     class Action:
         """
-        The actions that Auditlog distinguishes: creating, updating and deleting objects. Viewing objects is not logged.
-        The values of the actions are numeric, a higher integer value means a more intrusive action. This may be useful
-        in some cases when comparing actions because the ``__lt``, ``__lte``, ``__gt``, ``__gte`` lookup filters can be
-        used in queries.
+        The actions that Auditlog distinguishes: creating, updating and deleting objects. Viewing objects
+        is not logged. The values of the actions are numeric, a higher integer value means a more intrusive
+        action. This may be useful in some cases when comparing actions because the ``__lt``, ``__lte``,
+        ``__gt``, ``__gte`` lookup filters can be used in queries.
 
         The valid actions are :py:attr:`Action.CREATE`, :py:attr:`Action.UPDATE` and :py:attr:`Action.DELETE`.
         """
@@ -220,7 +268,9 @@ class LogEntry(models.Model):
     remote_addr = models.GenericIPAddressField(
         blank=True, null=True, verbose_name=_("remote address")
     )
-    timestamp = models.DateTimeField(auto_now_add=True, verbose_name=_("timestamp"))
+    timestamp = models.DateTimeField(
+        db_index=True, auto_now_add=True, verbose_name=_("timestamp")
+    )
     additional_data = models.JSONField(
         blank=True, null=True, verbose_name=_("additional data")
     )
@@ -258,9 +308,9 @@ class LogEntry(models.Model):
     @property
     def changes_str(self, colon=": ", arrow=" \u2192 ", separator="; "):
         """
-        Return the changes recorded in this log entry as a string. The formatting of the string can be customized by
-        setting alternate values for colon, arrow and separator. If the formatting is still not satisfying, please use
-        :py:func:`LogEntry.changes_dict` and format the string yourself.
+        Return the changes recorded in this log entry as a string. The formatting of the string can be
+        customized by setting alternate values for colon, arrow and separator. If the formatting is still
+        not satisfying, please use :py:func:`LogEntry.changes_dict` and format the string yourself.
 
         :param colon: The string to place between the field name and the values.
         :param arrow: The string to place between each old and new value.
@@ -303,7 +353,7 @@ class LogEntry(models.Model):
             values_display = []
             # handle choices fields and Postgres ArrayField to get human readable version
             choices_dict = None
-            if getattr(field, "choices") and len(field.choices) > 0:
+            if getattr(field, "choices", None) and len(field.choices) > 0:
                 choices_dict = dict(field.choices)
             if (
                 hasattr(field, "base_field")
@@ -327,7 +377,7 @@ class LogEntry(models.Model):
                             values_display.append(choices_dict.get(value, "None"))
                     except ValueError:
                         values_display.append(choices_dict.get(value, "None"))
-                    except:
+                    except Exception:
                         values_display.append(choices_dict.get(value, "None"))
             else:
                 try:
@@ -364,21 +414,22 @@ class LogEntry(models.Model):
 
 class AuditlogHistoryField(GenericRelation):
     """
-    A subclass of py:class:`django.contrib.contenttypes.fields.GenericRelation` that sets some default variables. This
-    makes it easier to access Auditlog's log entries, for example in templates.
+    A subclass of py:class:`django.contrib.contenttypes.fields.GenericRelation` that sets some default
+    variables. This makes it easier to access Auditlog's log entries, for example in templates.
 
-    By default this field will assume that your primary keys are numeric, simply because this is the most common case.
-    However, if you have a non-integer primary key, you can simply pass ``pk_indexable=False`` to the constructor, and
-    Auditlog will fall back to using a non-indexed text based field for this model.
+    By default this field will assume that your primary keys are numeric, simply because this is the most
+    common case. However, if you have a non-integer primary key, you can simply pass ``pk_indexable=False``
+    to the constructor, and Auditlog will fall back to using a non-indexed text based field for this model.
 
-    Using this field will not automatically register the model for automatic logging. This is done so you can be more
-    flexible with how you use this field.
+    Using this field will not automatically register the model for automatic logging. This is done so you
+    can be more flexible with how you use this field.
 
     :param pk_indexable: Whether the primary key for this model is not an :py:class:`int` or :py:class:`long`.
     :type pk_indexable: bool
-    :param delete_related: By default, including a generic relation into a model will cause all related objects to be
-        cascade-deleted when the parent object is deleted. Passing False to this overrides this behavior, retaining
-        the full auditlog history for the object. Defaults to True, because that's Django's default behavior.
+    :param delete_related: By default, including a generic relation into a model will cause all related
+        objects to be cascade-deleted when the parent object is deleted. Passing False to this overrides this
+        behavior, retaining the full auditlog history for the object. Defaults to True, because that's
+        Django's default behavior.
     :type delete_related: bool
     """
 
