@@ -1,28 +1,34 @@
 import datetime
+import itertools
 import json
+from unittest import mock
 
-import django
 from dateutil.tz import gettz
+from django.apps import apps
 from django.conf import settings
-from django.contrib import auth
+from django.contrib.admin.sites import AdminSite
+from django.contrib.auth import get_user_model
 from django.contrib.auth.models import AnonymousUser, User
-from django.core.exceptions import ValidationError
+from django.contrib.contenttypes.models import ContentType
 from django.db.models.signals import pre_save
-from django.http import HttpResponse
-from django.test import RequestFactory, TestCase
+from django.test import RequestFactory, TestCase, override_settings
 from django.utils import dateformat, formats, timezone
 
+from auditlog.admin import LogEntryAdmin
+from auditlog.context import set_actor
 from auditlog.diff import model_instance_diff
 from auditlog.middleware import AuditlogMiddleware
 from auditlog.models import LogEntry
-from auditlog.registry import auditlog
+from auditlog.registry import AuditlogModelRegistry, auditlog
 from auditlog_tests.models import (
     AdditionalDataIncludedModel,
     AltPrimaryKeyModel,
     CharfieldTextfieldModel,
     ChoicesFieldModel,
     DateTimeFieldModel,
+    JSONModel,
     ManyRelatedModel,
+    ManyRelatedOtherModel,
     NoDeleteHistoryModel,
     PostgresArrayFieldModel,
     ProxyModel,
@@ -38,7 +44,11 @@ from auditlog_tests.models import (
 
 class SimpleModelTest(TestCase):
     def setUp(self):
-        self.obj = SimpleModel.objects.create(text="I am not difficult.")
+        self.obj = self.make_object()
+        super().setUp()
+
+    def make_object(self):
+        return SimpleModel.objects.create(text="I am not difficult.")
 
     def test_create(self):
         """Creation is logged correctly."""
@@ -48,17 +58,14 @@ class SimpleModelTest(TestCase):
         # Check for log entries
         self.assertEqual(obj.history.count(), 1, msg="There is one log entry")
 
-        try:
-            history = obj.history.get()
-        except obj.history.DoesNotExist:
-            self.assertTrue(False, "Log entry exists")
-        else:
-            self.assertEqual(
-                history.action, LogEntry.Action.CREATE, msg="Action is 'CREATE'"
-            )
-            self.assertEqual(
-                history.object_repr, str(obj), msg="Representation is equal"
-            )
+        history = obj.history.get()
+        self.check_create_log_entry(obj, history)
+
+    def check_create_log_entry(self, obj, history):
+        self.assertEqual(
+            history.action, LogEntry.Action.CREATE, msg="Action is 'CREATE'"
+        )
+        self.assertEqual(history.object_repr, str(obj), msg="Representation is equal")
 
     def test_update(self):
         """Updates are logged correctly."""
@@ -66,8 +73,7 @@ class SimpleModelTest(TestCase):
         obj = self.obj
 
         # Change something
-        obj.boolean = True
-        obj.save()
+        self.update(obj)
 
         # Check for log entries
         self.assertEqual(
@@ -77,7 +83,13 @@ class SimpleModelTest(TestCase):
         )
 
         history = obj.history.get(action=LogEntry.Action.UPDATE)
+        self.check_update_log_entry(obj, history)
 
+    def update(self, obj):
+        obj.boolean = True
+        obj.save()
+
+    def check_update_log_entry(self, obj, history):
         self.assertJSONEqual(
             history.changes,
             '{"boolean": ["False", "True"]}',
@@ -96,13 +108,17 @@ class SimpleModelTest(TestCase):
         self.assertJSONEqual(
             obj.history.get(action=LogEntry.Action.UPDATE).changes,
             '{"boolean": ["False", "True"]}',
-            msg="Object modifications that are not saved to DB are not logged when using the `update_fields`.",
+            msg=(
+                "Object modifications that are not saved to DB are not logged "
+                "when using the `update_fields`."
+            ),
         )
 
     def test_django_update_fields_edge_cases(self):
         """
-        The test ensures that if Django's `update_fields` behavior ever changes for special values `(None, [])`, the
-        package should too. https://docs.djangoproject.com/en/3.2/ref/models/instances/#specifying-which-fields-to-save
+        The test ensures that if Django's `update_fields` behavior ever changes for special
+        values `(None, [])`, the package should too.
+        https://docs.djangoproject.com/en/3.2/ref/models/instances/#specifying-which-fields-to-save
         """
         obj = self.obj
 
@@ -132,25 +148,27 @@ class SimpleModelTest(TestCase):
         """Deletion is logged correctly."""
         # Get the object to work with
         obj = self.obj
-
-        history = obj.history.latest()
+        content_type = ContentType.objects.get_for_model(obj.__class__)
+        pk = obj.pk
 
         # Delete the object
-        obj.delete()
+        self.delete(obj)
 
         # Check for log entries
-        self.assertEqual(
-            LogEntry.objects.filter(
-                content_type=history.content_type,
-                object_pk=history.object_pk,
-                action=LogEntry.Action.DELETE,
-            ).count(),
-            1,
-            msg="There is one log entry for 'DELETE'",
-        )
+        qs = LogEntry.objects.filter(content_type=content_type, object_pk=pk)
+        self.assertEqual(qs.count(), 1, msg="There is one log entry for 'DELETE'")
+
+        history = qs.get()
+        self.check_delete_log_entry(obj, history)
+
+    def delete(self, obj):
+        obj.delete()
+
+    def check_delete_log_entry(self, obj, history):
+        pass
 
     def test_recreate(self):
-        SimpleModel.objects.all().delete()
+        self.obj.delete()
         self.setUp()
         self.test_create()
 
@@ -173,16 +191,79 @@ class SimpleModelTest(TestCase):
         )  # must be created in default database
 
 
-class AltPrimaryKeyModelTest(SimpleModelTest):
+class NoActorMixin:
+    def check_create_log_entry(self, obj, log_entry):
+        super().check_create_log_entry(obj, log_entry)
+        self.assertIsNone(log_entry.actor)
+
+    def check_update_log_entry(self, obj, log_entry):
+        super().check_update_log_entry(obj, log_entry)
+        self.assertIsNone(log_entry.actor)
+
+    def check_delete_log_entry(self, obj, log_entry):
+        super().check_delete_log_entry(obj, log_entry)
+        self.assertIsNone(log_entry.actor)
+
+
+class WithActorMixin:
+    sequence = itertools.count()
+
     def setUp(self):
-        self.obj = AltPrimaryKeyModel.objects.create(
+        username = f"actor_{next(self.sequence)}"
+        self.user = get_user_model().objects.create(
+            username=username,
+            email=f"{username}@example.com",
+            password="secret",
+        )
+        super().setUp()
+
+    def tearDown(self):
+        self.user.delete()
+        super().tearDown()
+
+    def make_object(self):
+        with set_actor(self.user):
+            return super().make_object()
+
+    def check_create_log_entry(self, obj, log_entry):
+        super().check_create_log_entry(obj, log_entry)
+        self.assertEqual(log_entry.actor, self.user)
+
+    def update(self, obj):
+        with set_actor(self.user):
+            return super().update(obj)
+
+    def check_update_log_entry(self, obj, log_entry):
+        super().check_update_log_entry(obj, log_entry)
+        self.assertEqual(log_entry.actor, self.user)
+
+    def delete(self, obj):
+        with set_actor(self.user):
+            return super().delete(obj)
+
+    def check_delete_log_entry(self, obj, log_entry):
+        super().check_delete_log_entry(obj, log_entry)
+        self.assertEqual(log_entry.actor, self.user)
+
+
+class AltPrimaryKeyModelBase(SimpleModelTest):
+    def make_object(self):
+        return AltPrimaryKeyModel.objects.create(
             key=str(datetime.datetime.now()), text="I am strange."
         )
 
 
-class UUIDPrimaryKeyModelModelTest(SimpleModelTest):
-    def setUp(self):
-        self.obj = UUIDPrimaryKeyModel.objects.create(text="I am strange.")
+class AltPrimaryKeyModelTest(NoActorMixin, AltPrimaryKeyModelBase):
+    pass
+
+
+class AltPrimaryKeyModelWithActorTest(WithActorMixin, AltPrimaryKeyModelBase):
+    pass
+
+
+class UUIDPrimaryKeyModelModelBase(SimpleModelTest):
+    def make_object(self):
+        return UUIDPrimaryKeyModel.objects.create(text="I am strange.")
 
     def test_get_for_object(self):
         self.obj.boolean = True
@@ -200,29 +281,90 @@ class UUIDPrimaryKeyModelModelTest(SimpleModelTest):
         )
 
 
-class ProxyModelTest(SimpleModelTest):
-    def setUp(self):
-        self.obj = ProxyModel.objects.create(text="I am not what you think.")
+class UUIDPrimaryKeyModelModelTest(NoActorMixin, UUIDPrimaryKeyModelModelBase):
+    pass
+
+
+class UUIDPrimaryKeyModelModelWithActorTest(
+    WithActorMixin, UUIDPrimaryKeyModelModelBase
+):
+    pass
+
+
+class ProxyModelBase(SimpleModelTest):
+    def make_object(self):
+        return ProxyModel.objects.create(text="I am not what you think.")
+
+
+class ProxyModelTest(NoActorMixin, ProxyModelBase):
+    pass
+
+
+class ProxyModelWithActorTest(WithActorMixin, ProxyModelBase):
+    pass
 
 
 class ManyRelatedModelTest(TestCase):
     """
-    Test the behaviour of a many-to-many relationship.
+    Test the behaviour of many-to-many relationships.
     """
 
     def setUp(self):
         self.obj = ManyRelatedModel.objects.create()
-        self.rel_obj = ManyRelatedModel.objects.create()
-        self.obj.related.add(self.rel_obj)
+        self.recursive = ManyRelatedModel.objects.create()
+        self.related = ManyRelatedOtherModel.objects.create()
+        self.base_log_entry_count = (
+            LogEntry.objects.count()
+        )  # created by the create() calls above
 
-    def test_related(self):
+    def test_recursive(self):
+        self.obj.recursive.add(self.recursive)
         self.assertEqual(
-            LogEntry.objects.get_for_objects(self.obj.related.all()).count(),
-            self.rel_obj.history.count(),
+            LogEntry.objects.get_for_objects(self.obj.recursive.all()).first(),
+            self.recursive.history.first(),
         )
+
+    def test_related_add_from_first_side(self):
+        self.obj.related.add(self.related)
         self.assertEqual(
             LogEntry.objects.get_for_objects(self.obj.related.all()).first(),
-            self.rel_obj.history.first(),
+            self.related.history.first(),
+        )
+        self.assertEqual(LogEntry.objects.count(), self.base_log_entry_count + 1)
+
+    def test_related_add_from_other_side(self):
+        self.related.related.add(self.obj)
+        self.assertEqual(
+            LogEntry.objects.get_for_objects(self.obj.related.all()).first(),
+            self.related.history.first(),
+        )
+        self.assertEqual(LogEntry.objects.count(), self.base_log_entry_count + 1)
+
+    def test_related_remove_from_first_side(self):
+        self.obj.related.add(self.related)
+        self.obj.related.remove(self.related)
+        self.assertEqual(LogEntry.objects.count(), self.base_log_entry_count + 2)
+
+    def test_related_remove_from_other_side(self):
+        self.related.related.add(self.obj)
+        self.related.related.remove(self.obj)
+        self.assertEqual(LogEntry.objects.count(), self.base_log_entry_count + 2)
+
+    def test_related_clear_from_first_side(self):
+        self.obj.related.add(self.related)
+        self.obj.related.clear()
+        self.assertEqual(LogEntry.objects.count(), self.base_log_entry_count + 2)
+
+    def test_related_clear_from_other_side(self):
+        self.related.related.add(self.obj)
+        self.related.related.clear()
+        self.assertEqual(LogEntry.objects.count(), self.base_log_entry_count + 2)
+
+    def test_additional_data(self):
+        self.obj.related.add(self.related)
+        log_entry = self.obj.history.first()
+        self.assertEqual(
+            log_entry.additional_data, {"related_model_id": self.related.id}
         )
 
 
@@ -232,75 +374,66 @@ class MiddlewareTest(TestCase):
     """
 
     def setUp(self):
-        def get_response(request):
-            return HttpResponse()
-
-        self.middleware = AuditlogMiddleware(get_response)
+        self.get_response_mock = mock.Mock()
+        self.response_mock = mock.Mock()
+        self.middleware = AuditlogMiddleware(get_response=self.get_response_mock)
         self.factory = RequestFactory()
         self.user = User.objects.create_user(
             username="test", email="test@example.com", password="top_secret"
         )
 
+    def side_effect(self, assertion):
+        def inner(request):
+            assertion()
+            return self.response_mock
+
+        return inner
+
+    def assert_has_listeners(self):
+        self.assertTrue(pre_save.has_listeners(LogEntry))
+
+    def assert_no_listeners(self):
+        self.assertFalse(pre_save.has_listeners(LogEntry))
+
     def test_request_anonymous(self):
         """No actor will be logged when a user is not logged in."""
-        # Create a request
         request = self.factory.get("/")
         request.user = AnonymousUser()
 
-        # Run middleware
-        self.middleware.process_request(request)
+        self.get_response_mock.side_effect = self.side_effect(self.assert_no_listeners)
 
-        # Validate result
-        self.assertFalse(pre_save.has_listeners(LogEntry))
+        response = self.middleware(request)
 
-        # Finalize transaction
-        self.middleware.process_exception(request, None)
+        self.assertIs(response, self.response_mock)
+        self.get_response_mock.assert_called_once_with(request)
+        self.assert_no_listeners()
 
     def test_request(self):
         """The actor will be logged when a user is logged in."""
-        # Create a request
-        request = self.factory.get("/")
-        request.user = self.user
-        # Run middleware
-        self.middleware.process_request(request)
-
-        # Validate result
-        self.assertTrue(pre_save.has_listeners(LogEntry))
-
-        # Finalize transaction
-        self.middleware.process_exception(request, None)
-
-    def test_response(self):
-        """The signal will be disconnected when the request is processed."""
-        # Create a request
         request = self.factory.get("/")
         request.user = self.user
 
-        # Run middleware
-        self.middleware.process_request(request)
-        self.assertTrue(
-            pre_save.has_listeners(LogEntry)
-        )  # The signal should be present before trying to disconnect it.
-        self.middleware.process_response(request, HttpResponse())
+        self.get_response_mock.side_effect = self.side_effect(self.assert_has_listeners)
 
-        # Validate result
-        self.assertFalse(pre_save.has_listeners(LogEntry))
+        response = self.middleware(request)
+
+        self.assertIs(response, self.response_mock)
+        self.get_response_mock.assert_called_once_with(request)
+        self.assert_no_listeners()
 
     def test_exception(self):
         """The signal will be disconnected when an exception is raised."""
-        # Create a request
         request = self.factory.get("/")
         request.user = self.user
 
-        # Run middleware
-        self.middleware.process_request(request)
-        self.assertTrue(
-            pre_save.has_listeners(LogEntry)
-        )  # The signal should be present before trying to disconnect it.
-        self.middleware.process_exception(request, ValidationError("Test"))
+        SomeException = type("SomeException", (Exception,), {})
 
-        # Validate result
-        self.assertFalse(pre_save.has_listeners(LogEntry))
+        self.get_response_mock.side_effect = SomeException
+
+        with self.assertRaises(SomeException):
+            self.middleware(request)
+
+        self.assert_no_listeners()
 
 
 class SimpleIncludeModelTest(TestCase):
@@ -791,6 +924,173 @@ class UnregisterTest(TestCase):
         self.assertEqual(LogEntry.objects.count(), 0, msg="There are no log entries")
 
 
+class RegisterModelSettingsTest(TestCase):
+    def setUp(self):
+        self.test_auditlog = AuditlogModelRegistry()
+
+    def tearDown(self):
+        for model in self.test_auditlog.get_models():
+            self.test_auditlog.unregister(model)
+
+    def test_get_model_classes(self):
+        self.assertEqual(
+            len(list(self.test_auditlog._get_model_classes("auditlog"))),
+            len(list(apps.get_app_config("auditlog").get_models())),
+        )
+        self.assertEqual([], self.test_auditlog._get_model_classes("fake_model"))
+
+    def test_get_exclude_models(self):
+        # By default it returns DEFAULT_EXCLUDE_MODELS
+        self.assertEqual(len(self.test_auditlog._get_exclude_models(())), 2)
+
+        # Exclude just one model
+        self.assertTrue(
+            SimpleExcludeModel
+            in self.test_auditlog._get_exclude_models(
+                ("auditlog_tests.SimpleExcludeModel",)
+            )
+        )
+
+        # Exclude all model of an app
+        self.assertTrue(
+            SimpleExcludeModel
+            in self.test_auditlog._get_exclude_models(("auditlog_tests",))
+        )
+
+    def test_register_models_no_models(self):
+        self.test_auditlog._register_models(())
+
+        self.assertEqual(self.test_auditlog._registry, {})
+
+    def test_register_models_register_single_model(self):
+        self.test_auditlog._register_models(("auditlog_tests.SimpleExcludeModel",))
+
+        self.assertTrue(self.test_auditlog.contains(SimpleExcludeModel))
+        self.assertEqual(len(self.test_auditlog._registry), 1)
+
+    def test_register_models_register_app(self):
+        self.test_auditlog._register_models(("auditlog_tests",))
+
+        self.assertTrue(self.test_auditlog.contains(SimpleExcludeModel))
+        self.assertTrue(self.test_auditlog.contains(ChoicesFieldModel))
+        self.assertEqual(len(self.test_auditlog.get_models()), 19)
+
+    def test_register_models_register_model_with_attrs(self):
+        self.test_auditlog._register_models(
+            (
+                {
+                    "model": "auditlog_tests.SimpleExcludeModel",
+                    "include_fields": ["label"],
+                    "exclude_fields": [
+                        "text",
+                    ],
+                },
+            )
+        )
+
+        self.assertTrue(self.test_auditlog.contains(SimpleExcludeModel))
+        fields = self.test_auditlog.get_model_fields(SimpleExcludeModel)
+        self.assertEqual(fields["include_fields"], ["label"])
+        self.assertEqual(fields["exclude_fields"], ["text"])
+
+    def test_register_models_register_model_with_m2m_fields(self):
+        self.test_auditlog._register_models(
+            (
+                {
+                    "model": "auditlog_tests.ManyRelatedModel",
+                    "m2m_fields": {"related"},
+                },
+            )
+        )
+
+        self.assertTrue(self.test_auditlog.contains(ManyRelatedModel))
+        self.assertEqual(
+            self.test_auditlog._registry[ManyRelatedModel]["m2m_fields"], {"related"}
+        )
+
+    def test_register_from_settings_invalid_settings(self):
+        with override_settings(AUDITLOG_INCLUDE_ALL_MODELS="str"):
+            with self.assertRaisesMessage(
+                TypeError, "Setting 'AUDITLOG_INCLUDE_ALL_MODELS' must be a boolean"
+            ):
+                self.test_auditlog.register_from_settings()
+
+        with override_settings(AUDITLOG_EXCLUDE_TRACKING_MODELS="str"):
+            with self.assertRaisesMessage(
+                TypeError,
+                "Setting 'AUDITLOG_EXCLUDE_TRACKING_MODELS' must be a list or tuple",
+            ):
+                self.test_auditlog.register_from_settings()
+
+        with override_settings(AUDITLOG_EXCLUDE_TRACKING_MODELS=("app1.model1",)):
+            with self.assertRaisesMessage(
+                ValueError,
+                "In order to use setting 'AUDITLOG_EXCLUDE_TRACKING_MODELS', "
+                "setting 'AUDITLOG_INCLUDE_ALL_MODELS' must set to 'True'",
+            ):
+                self.test_auditlog.register_from_settings()
+
+        with override_settings(AUDITLOG_INCLUDE_TRACKING_MODELS="str"):
+            with self.assertRaisesMessage(
+                TypeError,
+                "Setting 'AUDITLOG_INCLUDE_TRACKING_MODELS' must be a list or tuple",
+            ):
+                self.test_auditlog.register_from_settings()
+
+        with override_settings(AUDITLOG_INCLUDE_TRACKING_MODELS=(1, 2)):
+            with self.assertRaisesMessage(
+                TypeError,
+                "Setting 'AUDITLOG_INCLUDE_TRACKING_MODELS' items must be str or dict",
+            ):
+                self.test_auditlog.register_from_settings()
+
+        with override_settings(AUDITLOG_INCLUDE_TRACKING_MODELS=({"test": "test"},)):
+            with self.assertRaisesMessage(
+                ValueError,
+                "Setting 'AUDITLOG_INCLUDE_TRACKING_MODELS' dict items must contain 'model' key",
+            ):
+                self.test_auditlog.register_from_settings()
+
+        with override_settings(AUDITLOG_INCLUDE_TRACKING_MODELS=({"model": "test"},)):
+            with self.assertRaisesMessage(
+                ValueError,
+                (
+                    "Setting 'AUDITLOG_INCLUDE_TRACKING_MODELS' model must be in the "
+                    "format <app_name>.<model_name>"
+                ),
+            ):
+                self.test_auditlog.register_from_settings()
+
+    @override_settings(
+        AUDITLOG_INCLUDE_ALL_MODELS=True,
+        AUDITLOG_EXCLUDE_TRACKING_MODELS=("auditlog_tests.SimpleExcludeModel",),
+    )
+    def test_register_from_settings_register_all_models_with_exclude_models(self):
+        self.test_auditlog.register_from_settings()
+
+        self.assertFalse(self.test_auditlog.contains(SimpleExcludeModel))
+        self.assertTrue(self.test_auditlog.contains(ChoicesFieldModel))
+
+    @override_settings(
+        AUDITLOG_INCLUDE_TRACKING_MODELS=(
+            {
+                "model": "auditlog_tests.SimpleExcludeModel",
+                "include_fields": ["label"],
+                "exclude_fields": [
+                    "text",
+                ],
+            },
+        )
+    )
+    def test_register_from_settings_register_models(self):
+        self.test_auditlog.register_from_settings()
+
+        self.assertTrue(self.test_auditlog.contains(SimpleExcludeModel))
+        fields = self.test_auditlog.get_model_fields(SimpleExcludeModel)
+        self.assertEqual(fields["include_fields"], ["label"])
+        self.assertEqual(fields["exclude_fields"], ["text"])
+
+
 class ChoicesFieldModelTest(TestCase):
     def setUp(self):
         self.obj = ChoicesFieldModel.objects.create(
@@ -831,6 +1131,12 @@ class ChoicesFieldModelTest(TestCase):
             msg="The human readable text 'Red' is displayed.",
         )
 
+    def test_changes_display_dict_many_to_one_relation(self):
+        obj = SimpleModel()
+        obj.save()
+        history = obj.history.get()
+        assert "related_models" in history.changes_display_dict
+
 
 class CharfieldTextfieldModelTest(TestCase):
     def setUp(self):
@@ -932,7 +1238,7 @@ class AdminPanelTest(TestCase):
         res = self.client.get("/admin/auditlog/logentry/")
         assert res.status_code == 200
         res = self.client.get("/admin/auditlog/logentry/add/")
-        assert res.status_code == 200
+        assert res.status_code == 403
         res = self.client.get(f"/admin/auditlog/logentry/{log_pk}/", follow=True)
         assert res.status_code == 200
         res = self.client.get(f"/admin/auditlog/logentry/{log_pk}/delete/")
@@ -941,6 +1247,89 @@ class AdminPanelTest(TestCase):
         assert res.status_code == 200
 
 
+class DiffMsgTest(TestCase):
+    def setUp(self):
+        super().setUp()
+        self.site = AdminSite()
+        self.admin = LogEntryAdmin(LogEntry, self.site)
+
+    def _create_log_entry(self, action, changes):
+        return LogEntry.objects.log_create(
+            SimpleModel.objects.create(),  # doesn't affect anything
+            action=action,
+            changes=json.dumps(changes),
+        )
+
+    def test_changes_msg__delete(self):
+        log_entry = self._create_log_entry(LogEntry.Action.DELETE, {})
+
+        self.assertEqual(self.admin.msg(log_entry), "")
+
+    def test_changes_msg__create(self):
+        log_entry = self._create_log_entry(
+            LogEntry.Action.CREATE,
+            {
+                "field two": [None, 11],
+                "field one": [None, "a value"],
+            },
+        )
+
+        self.assertEqual(
+            self.admin.msg(log_entry),
+            (
+                "<table>"
+                "<tr><th>#</th><th>Field</th><th>From</th><th>To</th></tr>"
+                "<tr><td>1</td><td>field one</td><td>None</td><td>a value</td></tr>"
+                "<tr><td>2</td><td>field two</td><td>None</td><td>11</td></tr>"
+                "</table>"
+            ),
+        )
+
+    def test_changes_msg__update(self):
+        log_entry = self._create_log_entry(
+            LogEntry.Action.UPDATE,
+            {
+                "field two": [11, 42],
+                "field one": ["old value of field one", "new value of field one"],
+            },
+        )
+
+        self.assertEqual(
+            self.admin.msg(log_entry),
+            (
+                "<table>"
+                "<tr><th>#</th><th>Field</th><th>From</th><th>To</th></tr>"
+                "<tr><td>1</td><td>field one</td><td>old value of field one</td>"
+                "<td>new value of field one</td></tr>"
+                "<tr><td>2</td><td>field two</td><td>11</td><td>42</td></tr>"
+                "</table>"
+            ),
+        )
+
+    def test_changes_msg__m2m(self):
+        log_entry = self._create_log_entry(
+            LogEntry.Action.UPDATE,
+            {  # mimicking the format used by log_m2m_changes
+                "some_m2m_field": {
+                    "type": "m2m",
+                    "operation": "add",
+                    "objects": ["Example User (user 1)", "Illustration (user 42)"],
+                },
+            },
+        )
+
+        self.assertEqual(
+            self.admin.msg(log_entry),
+            (
+                "<table>"
+                "<tr><th>#</th><th>Relationship</th><th>Action</th><th>Objects</th></tr>"
+                "<tr><td>1</td><td>some_m2m_field</td><td>add</td><td>Example User (user 1)"
+                "<br>Illustration (user 42)</td></tr>"
+                "</table>"
+            ),
+        )
+
+
 class NoDeleteHistoryTest(TestCase):
     def test_delete_related(self):
         instance = SimpleModel.objects.create(integer=1)
@@ -970,3 +1359,122 @@ class NoDeleteHistoryTest(TestCase):
             list(entries.values_list("action", flat=True)),
             [LogEntry.Action.CREATE, LogEntry.Action.UPDATE, LogEntry.Action.DELETE],
         )
+
+
+class JSONModelTest(TestCase):
+    def setUp(self):
+        self.obj = JSONModel.objects.create()
+
+    def test_update(self):
+        """Changes on a JSONField are logged correctly."""
+        # Get the object to work with
+        obj = self.obj
+
+        # Change something
+        obj.json = {
+            "quantity": "1",
+        }
+        obj.save()
+
+        # Check for log entries
+        self.assertEqual(
+            obj.history.filter(action=LogEntry.Action.UPDATE).count(),
+            1,
+            msg="There is one log entry for 'UPDATE'",
+        )
+
+        history = obj.history.get(action=LogEntry.Action.UPDATE)
+
+        self.assertJSONEqual(
+            history.changes,
+            '{"json": ["{}", "{\'quantity\': \'1\'}"]}',
+            msg="The change is correctly logged",
+        )
+
+    def test_update_with_no_changes(self):
+        """No changes are logged."""
+        first_json = {
+            "quantity": "1814",
+            "tax_rate": "17",
+            "unit_price": "144",
+            "description": "Method form.",
+            "discount_rate": "42",
+            "unit_of_measure": "bytes",
+        }
+        obj = JSONModel.objects.create(json=first_json)
+
+        # Change the order of the keys but not the values
+        second_json = {
+            "tax_rate": "17",
+            "description": "Method form.",
+            "quantity": "1814",
+            "unit_of_measure": "bytes",
+            "unit_price": "144",
+            "discount_rate": "42",
+        }
+        obj.json = second_json
+        obj.save()
+
+        # Check for log entries
+        self.assertEqual(
+            first_json,
+            second_json,
+            msg="dicts are the same",
+        )
+        self.assertEqual(
+            obj.history.filter(action=LogEntry.Action.UPDATE).count(),
+            0,
+            msg="There is no log entry",
+        )
+
+
+class ModelInstanceDiffTest(TestCase):
+    def test_diff_models_with_related_fields(self):
+        """No error is raised when comparing models with related fields."""
+
+        # This tests that track_field() does indeed ignore related fields.
+
+        # a model without reverse relations
+        simple1 = SimpleModel()
+        simple1.save()
+
+        # a model with reverse relations
+        simple2 = SimpleModel()
+        simple2.save()
+        related = RelatedModel(related=simple2, one_to_one=simple2)
+        related.save()
+
+        # Demonstrate that simple1 can have DoesNotExist on reverse
+        # OneToOne relation.
+        with self.assertRaises(
+            SimpleModel.reverse_one_to_one.RelatedObjectDoesNotExist
+        ):
+            simple1.reverse_one_to_one  # equals None
+
+        # accessing relatedmodel_set won't trigger DoesNotExist.
+        self.assertEqual(simple1.related_models.count(), 0)
+
+        # simple2 DOES have these relations
+        self.assertEqual(simple2.reverse_one_to_one, related)
+        self.assertEqual(simple2.related_models.count(), 1)
+
+        model_instance_diff(simple2, simple1)
+        model_instance_diff(simple1, simple2)
+
+    def test_when_field_doesnt_exist(self):
+        """No error is raised and the default is returned."""
+        first = SimpleModel(boolean=True)
+        second = SimpleModel()
+
+        # then boolean should be False, as we use the default value
+        # specified inside the model
+        del second.boolean
+
+        changes = model_instance_diff(first, second)
+
+        # Check for log entries
+        self.assertEqual(
+            changes,
+            {"boolean": ("True", "False")},
+            msg="ObjectDoesNotExist should be handled",
+        )
